# -*- coding: utf-8 -*-
"""
Lumerical functions - How best to generate objects in python.
Every single object function must return a dict which details their x-y-z coordinate, their bounding box normalised to x-y coordinates), object rotation etc. 
We note that the objects cannot be modified FROM this unless lumemrical outputs an object control struct, but we'll see about that
Created on Thu Nov 13 13:36:03 202
@author: vidar
"""
import numpy as np
from scipy import constants
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import lumapi
import os
import numpy as np
#%%

def select_and_set_props(sim, name, propdict):
    """
    If name=None, we assume it's already been selected in the scope, so you can use this after creation without passing a name
    """
    if name:
        sim.select(name)
    for key, item in propdict.items():    
        sim.set(key,item)

def create_groups_from_dict(sim, grouplist):
    """
    Create structure and analysis groups in Lumerical without checking for existence.
    Uses temporary SETTER groups and Lumerical's addtogroup behavior.
    Only maximal (non-subset) paths are used.
    
    grouplist example:
    {
        "structure": ["Geometry", "Geometry::Substrate", "Geometry::SurfaceComponents"],
        "analysis": ["Analysis::DFTMonitors","Analysis::VideoMonitors"]
    }
    """
    
    def filter_maximal_paths(paths):
        """Keep only paths that are not prefixes of other paths"""
        paths_sorted = sorted(paths, key=lambda x: -len(x))  # longest first
        maximal = []
        for p in paths_sorted:
            if not any(other.startswith(p + "::") for other in maximal):
                maximal.append(p)
        return maximal
    
    # --- Structure groups ---
    struct_paths = filter_maximal_paths(grouplist.get("structure", []))
    for path in struct_paths:
        sim.addstructuregroup()
        sim.set("name", "structure_setter")
        sim.addtogroup(path)
        sim.delete()
    
    # --- Analysis groups ---
    analysis_paths = filter_maximal_paths(grouplist.get("analysis", []))
    
    if analysis_paths:
        # Find all unique top-level analysis groups
        top_level_analysis = set(p.split("::")[0] for p in analysis_paths)
        
        # Create each top-level analysis group
        for top_group in top_level_analysis:
            sim.addanalysisgroup()
            sim.set("name", top_group)
        
        # Add all paths under their respective top-level analysis groups
        for path in analysis_paths:
            top_group = path.split("::")[0]
            sim.addanalysisgroup()
            sim.set("name", "analysis_setter")
            sim.addtogroup(path)
            sim.delete()
            
def are_all_dict_values_type(d,ttype=None):
    if isinstance(d, dict):  # If d is a dictionary, check all values
        return all(are_all_dict_values_type(value) for value in d.values())
    elif isinstance(d, list):  # If d is a list, check all elements
        return all(are_all_dict_values_type(value) for value in d)
    else:  # If it's neither a dict nor a list, just check if the value is None
        return d is None


class RoundedCuboid:
    def __init__(self, rx=0, ry=0, rz=0, Dx=1, Dy=1, Dz=1, rx2=None, ry2=None, rz2=None):
        """
        x1,y2,z2 ----> o--------o <- x2,y2,z2
                      /        /|                         xy2
                     /        / |                          ^
        x1,y1,z2 -> o--------o  |                          I
                    |  o-----|  o <- x2,y2,z1     yz1<- |     | ->yz2  (out of cube = xz1, into cube xz2)
                    | /      | /                          I
                    |/       |/                           v
        x1,y1,z1--> o--------o <---- x2,y1,z1            xy1
                    
                    <---Dx--->
        
        Rounding Radius Explanation:
            
        Rounding radius can be different for the normal axis of planes with shared normals. So the zy plane can have different rx values, but not different ry and rz values.
        To make this easier, we will run rounding as a "per exception": rx/ry/rz are FOR ALL, and then you can elect to set rx2/ry2/rz2
        Examples: 
            I want rx1 == rx2: set rx = value, and ignore rx2
            I want rx1 = val, and rx2 = 0: Set rx = value, and rx2=0
            I want rx1 = 0.5 and rx2 = 1, you just set rx=0.5 and rx2=1.  
            
        This way, we can also remove planes completely for FLAT faces (if you want to make plus symbols or something)
        
        note that x1 = -1, anx x2 = +1 and vice versa!        

        Dx/Dy/Dz Explanation:
        This represents the TOTAL bounds of the object. This means that the rounding radius of any corner cannot exceed D/2, since then 2r == D
        If greater, then the sphere will clip out of the opposite side. Note, this also means that rx1 can be small, while rx2 can be D/2.
        
        We build objects by doing:
            rx1 -> Dxp -> rx2 (where Dxp = Dx-(rx1+rx2))
            This naturally also means that rx1+rx2 !> Dx
            But you can have: Dx = 100, rx1 = 5, Dxp = 45, rx2 = 50 and the maths will still work out! 
        """
        
        #Fix rx/ry/rz overlap, maximum permitted is D/2. After this, you might as well make a custom shape from spheres instead.
        
        #Assign new r2 values if not assigned, then fix the r_overlap. We will do a rudimentary check to see if rounding can be ratiod, since if r1 = 1.5r2, then r1 should equal D/2, and r2 = D/(2*1.5) 
        r1_list  = [rx,  ry,  rz]
        r2_list  = [rx2, ry2, rz2]
        D_list   = [Dx,  Dy,  Dz]
        self.rr = {}
        for i, axis in enumerate("xyz"):
            self.rr[axis] = list(self._fix_overlap_and_r2(r1_list[i], r2_list[i], D_list[i]))
        
        self.base_corners = {}
        self.base_edges   = {}
        
        self.corner_props = {}
        self.edge_props = {}
        self.core_props = {}
        
        self.Dx = Dx; self.Dy = Dy; self.Dz = Dz
        
        #Get base locations of all corners, generating the keys for future use
        self._get_base_corners()
        #Get real locations of all corners and rounding radiuses.
        self._get_corner_props()
        
        #Get the base keys for edges, and generate keys for future use. Remember to use frozenset() when calling edge keys.
        self._get_base_edges() 
        
        self._get_edge_props()
        
        self._get_core_props()
        
        self._generate_primitives()

    def _fix_overlap_and_r2(self, r1, r2, D):
        # 1. assign r2 if missing
        if r2 is None:
            r2 = r1
        r_lim = D / 2
    
        # 2. if nothing exceeds the limit, return unchanged
        if r1 <= r_lim and r2 <= r_lim:
            return r1, r2
    
        # 3. ratio k = min / max
        radii = [r1, r2]
        rmin = min(radii); rmax = max(radii)
        k = rmin / rmax if rmax != 0 else 0   # safe even if both 0
    
        # 4. index positions of min and max in original order
        idx_min = radii.index(rmin); idx_max = radii.index(rmax)
    
        # 5. build new radii in correct original order
        new = [None, None]
        new[idx_max] = r_lim
        new[idx_min] = k * r_lim
    
        return new[0], new[1]
    
    #CODE FOR BASE PROPERTIES, THESE WILL ALWYS BE THE SAME FOR ALL RESULTS!
    #~~~~~~~~~~~~~~~~ START ~~~~~~~~~~~~~~~~#
    
    def _get_base_corners(self):
        for xi,x_coord in enumerate((-1, 1)):
            for yi,y_coord in enumerate((-1, 1)):
                for zi,z_coord in enumerate((-1, 1)):
                    self.base_corners[(x_coord, y_coord, z_coord)] = {"rr":{"rx": self.rr["x"][xi], "ry": self.rr["y"][yi], "rz": self.rr["z"][zi]},
                                                                       "loc":{"x":x_coord,"y":y_coord,"z":z_coord}}
    def _get_base_edges(self):
        corners = list(self.corner_props.keys())
        edges = []
        
        for i, c1 in enumerate(corners):
            for j in range(i+1, len(corners)):
                c2 = corners[j]
                diffs = sum(a != b for a,b in zip(c1, c2))
                if diffs == 1:
                    edges.append((c1, c2))
        self.base_edges = edges
    #~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~#
    
    #CODE FOR REAL CORNER, EDGE AND CORE PROPS#
    #~~~~~~~~~~~~~~~~ START ~~~~~~~~~~~~~~~~#
    def _get_corner_props(self):
        for xi,yi,zi in self.base_corners.keys():
            
            x = xi*self.Dx/2 - xi*self.base_corners[(xi,yi,zi)]["rr"]["rx"]
            y = yi*self.Dy/2 - yi*self.base_corners[(xi,yi,zi)]["rr"]["ry"]
            z = zi*self.Dz/2 - zi*self.base_corners[(xi,yi,zi)]["rr"]["rz"]
            
            
            self.corner_props[(xi,yi,zi)] = {"rr":{"x": self.base_corners[(xi,yi,zi)]["rr"]["rx"], "y": self.base_corners[(xi,yi,zi)]["rr"]["ry"], "z": self.base_corners[(xi,yi,zi)]["rr"]["rz"]},             
                                             "loc":{"x":x,"y":y,"z":z}}

    def _get_edge_props(self):
        self.edge_props = {}
    
        for c1, c2 in self.base_edges:
            loc1 = self.corner_props[c1]["loc"].copy()
            loc2 = self.corner_props[c2]["loc"].copy()
    
            # Determine the edge axis (where the corner coordinates differ)
            diffs = [a != b for a, b in zip(c1, c2)]
            normal_axis = ["x", "y", "z"][diffs.index(True)]
    
            # Get the in-plane radii from one of the corners (they should be equal)
            rr_corner = self.corner_props[c1]["rr"].copy()
            rr_edge = {}
            for axis in "xyz":
                if axis == normal_axis:
                    rr_edge[axis] = None
                else:
                    rr_edge[axis] = rr_corner[axis]
    
            # Store in edge_props
            self.edge_props[(c1, c2)] = {
                "loc": [[loc1["x"], loc1["y"], loc1["z"]],
                        [loc2["x"], loc2["y"], loc2["z"]]],
                "rr": rr_edge}
    
    def _get_core_props(self):
        #We assume that each rounded cube has 3 core cubes, and these will lie along the x/y/z normals. 
        #This is actually trivial, since all cubes are extruded from the core cube that we can extract from all corner radii. 
        dim_mod = {"x":{"val":self.Dx,"ind":0},"y":{"val":self.Dy,"ind":1},"z":{"val":self.Dz,"ind":2}}
        for norm in ["x","y","z"]:
            self.core_props[norm] = {}
            for corner,val in self.corner_props.items():                
                self.core_props[norm][corner] = {"loc":val["loc"].copy()}
                
                self.core_props[norm][corner]["loc"][norm] = corner[dim_mod[norm]["ind"]]*dim_mod[norm]["val"]/2
               
    #~~~~~~~~~~~~~~~~ END ~~~~~~~~~~~~~~~~#
    
    def _generate_primitives(self):
        self.r_spheres = []
        self.r_cylinders = []
        self.c_cubes = []
        
        
        # ----------- ROUNDING SPHERES -----------
         
        for corner, props in self.corner_props.items():
            rr = props["rr"]
            if any(rr[ax] == 0 for ax in "xyz"):
                continue
            entry = {"loc": props["loc"].copy(), "radius": rr.copy()}
            if entry not in self.r_spheres:
                self.r_spheres.append(entry)
        
        
       # ----------- CYLINDERS -----------
        for edge, props in self.edge_props.items():
            loc0, loc1 = props["loc"]
            rr = props["rr"]
        
            # skip zero-length edges
            if loc0 == loc1:
                continue
        
            # find cylinder axis (where radius is None)
            norm = next(ax for ax, v in rr.items() if v is None)
        
            # skip if any non-axis radius is zero
            if any(rr[ax] == 0 for ax in "xyz" if ax != norm):
                continue
        
            # compute cylinder center
            center = {ax: 0.5 * (loc0[i] + loc1[i]) for i, ax in enumerate("xyz")}
        
            # compute cylinder height along norm axis
            height = abs(loc1["xyz".index(norm)] - loc0["xyz".index(norm)])
        
            # perpendicular radii
            radius = {ax: rr[ax] for ax in "xyz" if ax != norm}
        
            entry = {
                "norm": norm,
                "loc": center,
                "radius": radius,
                "height": height
            }
        
            # avoid duplicates
            if entry not in self.r_cylinders:
                self.r_cylinders.append(entry)
    
        # ----------- CORE CUBES -----------
        for norm, cubes in self.core_props.items():
            cube_range = {}
            degenerate = False
            
            for ax in "xyz":
                # grab all loc[ax] for all corners in this norm
                vals = [props["loc"][ax] for props in cubes.values()]
                lo, hi = min(vals), max(vals)
                
                if lo == hi:
                    degenerate = True
                    break
                
                cube_range[f"{ax}min"] = lo
                cube_range[f"{ax}max"] = hi
        
            if not degenerate:
                self.c_cubes.append({"range": cube_range})

                
    def __repr__(self):
        return f"RoundedCuboidRadii({self.corner_props})"


class Nanowire:
    def __init__(self,
        radius,
        length,
        shape="circle",
        endcaps="both",          # "none", "top", "bottom", "both"
        cap_factor=0.5,          # squash factor for z-radius
        seed="none",             # "none", "top", "bottom", "both"
        seed_rfactor = 0.1,      # Percentage of radius to offset particle by. 0.1 = 0.1r, -0.1 = -0.1r
        seed_z_offset=0.1,):
        
        self.radius   = radius
        self.length   = length
                
        self.endcaps = endcaps
        self.cap_factor = cap_factor
        
        self.seed = seed
        self.seed_rfactor = seed_rfactor
        self.shape = shape
        
        # Storage (mirrors your RoundedCuboid structure)
        self.core_cylinder = []
        self.endcaps_list = []
        self.seed_list = []
        
        # Build geometry
        self._make_cylinder()
        self._make_endcaps()
        self._make_seeds()
    
    # -----------------------------------------------------------
    #  MAIN CYLINDER (your cylinder dict format)
    # -----------------------------------------------------------
    def _make_cylinder(self):
        cyl = {
            "loc": {"x": 0, "y": 0, "z": self.length/2},
            "radius": {"x": self.radius, "y": self.radius, "z": self.length/2},
            "range": {"zmin": -self.length/2, "zmax": self.length/2},
            "norm": [0, 0, 1] #Technically, we cannot choose this at all...
        }
        self.core_cylinder.append(cyl)
    
    # -----------------------------------------------------------
    #  ENDCAPS (deformed spheres) Note that this is only supported for shape=circle!
    # -----------------------------------------------------------
    def _make_endcaps(self):

        if self.endcaps not in ("bottom", "top", "both") or self.shape!="circle":
            return
        
        r = self.radius
        rz = r * self.cap_factor
        
        # Bottom
        if self.endcaps in ("bottom", "both"):
            cap = {
                "loc": {"x": 0, "y": 0, "z": -self.length/2},
                "radius": {"x": r, "y": r, "z": rz}
            }
            self.endcaps_list.append(cap)
        
        # Top
        if self.endcaps in ("top", "both"):
            cap = {
                "loc": {"x": 0, "y": 0, "z": self.length/2},
                "radius": {"x": r, "y": r, "z": rz}
            }
            self.endcaps_list.append(cap)
    
    
    # -----------------------------------------------------------
    #  SEEDS (small spheres with different material)
    # We need to make both, but we set the z-order of the "air" sphere to be at the bottom of everything.
    # -----------------------------------------------------------
    def _make_seeds(self):
        if self.seed not in ("bottom", "top", "both"):
            return
        
        seed_rz = self.radius * self.seed_rfactor
        
        # Bottom seed
        if self.seed in ("bottom", "both"):
            s = {
                "loc": {"x": 0, "y": 0, "z": -self.length/2 - self.seed_rfactor*self.radius},
                "radius": {"x": seed_rz, "y": seed_rz, "z": seed_rz}
            }
            self.seed_list.append(s)
        
        # Top seed
        if self.seed in ("top", "both"):
            s = {
                "loc": {"x": 0, "y": 0, "z": self.length/2 + self.seed_rfactor*self.radius},
                "radius": {"x": seed_rz, "y": seed_rz, "z": seed_rz},
            }
            self.seed_list.append(s)

def coordinate_standardisation(method = "span", x=None,y=None,z=None,Dx=None,Dy=None,Dz=None,rx=None,ry=None,rz=None,xmm=None,ymm=None,zmm=None):
    """
    This function works by taking a limited set of co-ordinates, and outputting the full set.
    This means that if you provide: x,y,z and Dx,Dy,Rz, it will give you xyz = [x,y,z], Dxyz = [Dx,Dy,Dz], rxyz = [rx,ry,rz], mmxyz = [xmm,ymm,zmm]
    How does it solve redundancy? Well, if you give it Dx and xmm, these will compete for space.
    The solution is to always prefer Dx/rx outputs over min-max, unless not provided...
    
    If the MODE is set to "span", then we ignore mmxyz until the end, and calculate Dx/rx from whichever is not None. 
    If both are provided, then we will raise an error - but continue as intended. Since if a sphere, rx will be used instead of Dx unless it doesn't exist, at which point it will be calculated 
    if the MODE is set to "minmax", then we calculate Dx from mm
    
    What this func needs to cover:
        x = None, and xmm is not -> calculate x from xmm
        x = value, and xmm range does not match -> recalculate x from xmm
    """    
    xyz    = [x,y,z]
    Dxyz   = [Dx,Dy,Dz]
    rxyz   = [rx,ry,rz] 
    mmxyz  = [xmm,ymm,zmm]
    ## First check if this is a default generation, and produce defaults in that case.
    for i,vals in enumerate(zip(xyz,Dxyz,rxyz,mmxyz)):
        coord, D, r, mm = vals
        if all(a == None for a in [coord,D,r,mm]):
            xyz[i] = 0
            Dxyz[i] = 1e-6
            rxyz[i]  = Dxyz[i]/2 
            mmxyz[i] = [-rxyz[i],rxyz[i]]
        
        if method == "span":
            #Check if all span entries are none
            if all (a == None for a in [D,r]):
                #use min-max fallback, if it exists EVEN if mode is set to span.
                if mmxyz[i] != None:
                    Dxyz[i] = mmxyz[i][1] - mmxyz[i][0]; rxyz[i] = Dxyz[i]/2
                    xyz[i] = np.mean(mmxyz[i])
                #Else, set defaults for all values!
                else:
                    Dxyz[i] = 1e-6; rxyz[i] = Dxyz[i]/2; 
                    if coord == None:
                        xyz[i] = 0
                    mmxyz[i] = [xyz[i]-rxyz[i],xyz[i]+rxyz[i]]
            else:        
                #Since not all values are none, we check first if r is none, and then alter r and D accordingly.
                if r == None: rxyz[i] = D/2
                if D == None: Dxyz[i] = r*2
                if coord == None: xyz[i] = 0
                #And now we alter the minmax values to match the mode
                mmxyz[i] = [xyz[i]-rxyz[i],xyz[i]+rxyz[i]]
                
        if method == "minmax":
            # Check if min-max is none
            if mm == None:
                # attempt to use span fallback
                if not all(a == None for a in [D, r]):
                    # prefer rx, since this is closer analogue to minmax and will work on spheres too
                    if r == None: rxyz[i] = D/2
                    if D == None: Dxyz[i] = r*2
                    if coord == None: xyz[i] = 0

                    mmxyz[i] = [xyz[i] - rxyz[i], xyz[i] + rxyz[i]]

                # Else, set defaults
                else:
                    Dxyz[i] = 1e-6; rxyz[i] = Dxyz[i]/2; 
                    if coord == None:
                        xyz[i] = 0
                    mmxyz[i] = [xyz[i]-rxyz[i],xyz[i]+rxyz[i]]

            else:
                # min-max single-handedly otherwise governs all parameter spaces
                mmxyz[i] = mm
                xyz[i]  = np.mean(mm)
                Dxyz[i] = mm[1] - mm[0]
                rxyz[i] = Dxyz[i] / 2
    return(xyz,Dxyz,rxyz,mmxyz)
                
        
    
def L_primitive(sim,primitive="rect", method="span",x=0,y=0,z=0, Dx=None,rx=None,Dy=None,ry=None,Dz=None,rz=None,norm="z",xminmax = [0,0], yminmax=[0,0],zminmax = [0,0],material=None,zorder=0):
    
    #First we handle the alias case, since we don't want to run the function with illegitimate values.
    #In this case, we want all primitives to have the same reasonable dimension of 1e-6 (where Dx = 2rx)
    xyz,D,r,mm = coordinate_standardisation(method = method, x=x,y=y,z=z,Dx=Dx,Dy=Dy,Dz=Dz,rx=rx,ry=ry,rz=rz,xmm=xminmax,ymm=yminmax,zmm=zminmax)
    axstr = ["x","y","z"]
        
    if primitive == "sphere":
        
        name = sim.addsphere()
        for i,axs in enumerate(axstr):
            sim.setnamed(name, axs, xyz[i])
            sim.setnamed(name, "radius "+axs, r[i])

        if material: sim.setnamed(name, "material", material)
        sim.setnamed(name, "z order", zorder)
    
    if primitive == "rect":
        name = sim.addsphere()
        for i,axs in enumerate(axstr):
            sim.setnamed(name, axs, xyz[i])
            sim.setnamed(name, axs+" span", D[i])
        if material: sim.setnamed(name, "material", material)
        sim.setnamed(name, "z order", zorder)
    
    if primitive == "cylinder":
        name = sim.addcylinder()
        # Default orientation: along z
        # axis_map defines which axis is the primary cylinder axis
        axis_map = {"x": "X", "y": "Y", "z": "Z"}
        rot_map  = {"x": [0, 1, 0], "y": [1, 0, 0], "z": [0, 0, 1]}
        
        primary_axis = axis_map.get(norm.lower(), "Z")  # fallback to z if unknown
        rotation     = rot_map.get(norm.lower(), [0, 0, 1])
        
        # Set cylinder center
        for i, axs in enumerate(["x", "y", "z"]):
            sim.setnamed(name, axs, xyz[i])
        
        # Set cylinder radius and height
        # Convention: radius = r perpendicular to axis, height = D along axis
        # Map: primary axis gets D, other axes use radius
        for i, axs in enumerate(["x", "y", "z"]):
            if axs.lower() == norm.lower():
                sim.setnamed(name, "height " + axs, D[i])
            else:
                sim.setnamed(name, "radius " + axs, r[i])
        
        # Set rotation
        sim.setnamed(name, "primary axis", primary_axis)
        sim.setnamed(name, "rotation", rotation)
        
        if material: sim.setnamed(name, "material", material)
        sim.setnamed(name, "z order", zorder)


def get_minmax_items(xmm,ymm,zmm):
    return([["x","y","z"],["min","max"],[xmm,ymm,zmm]])

def rot_matrix(xr, yr, zr):
    cx, sx = np.cos(np.radians(xr)), np.sin(np.radians(xr))
    cy, sy = np.cos(np.radians(yr)), np.sin(np.radians(yr))
    cz, sz = np.cos(np.radians(zr)), np.sin(np.radians(zr))

    Rx = np.array([[1, 0, 0],
                   [0, cx, -sx],
                   [0, sx,  cx]])

    Ry = np.array([[ cy, 0, sy],
                   [  0, 1,  0],
                   [-sy, 0, cy]])

    Rz = np.array([[cz, -sz, 0],
                   [sz,  cz, 0],
                   [ 0,   0, 1]])

    # Lumerical uses Rz * Ry * Rx
    return Rz @ Ry @ Rx

def aabb_of_rotated_cylinder(center, rx, ry, length, xr, yr, zr):
    cx, cy, cz = center
    h = length / 2

    # Local extreme points
    pts = np.array([
        [ rx,  0,  0],
        [-rx,  0,  0],
        [  0, ry,  0],
        [  0,-ry,  0],
        [  0,  0,  h],
        [  0,  0, -h],
    ])

    R = rot_matrix(xr, yr, zr)

    # Rotate + translate
    world_pts = (R @ pts.T).T + np.array(center)

    xmin, ymin, zmin = world_pts.min(axis=0)
    xmax, ymax, zmax = world_pts.max(axis=0)

    return({"rng":{"x":[xmin, xmax], "y":[ymin, ymax], "z":[zmin, zmax]}})

def quick_sphere(sim, xyz, rx, ry, rz, zorder=0, material=None,name=None):
    sim.addsphere()
    if name:
        sim.set("name",name)
    x,y,z = xyz
    sim.set("x", x)
    sim.set("y", y)
    sim.set("z", z)
    sim.set("radius", rx)
    sim.set("radius 2", ry)
    sim.set("radius 3", rz)
    if material: sim.set("material", material)
    sim.set("override mesh order from material database",1); 
    sim.set("mesh order",zorder);
    

def quick_cuboid(sim, xyz, xmm,ymm,zmm, zorder=0, material=None,name=None):
    sim.addrect()
    if name:
        sim.set("name",name)
    axnorm,mim,xyzmm = get_minmax_items(xmm,ymm,zmm) 
    for i,norm in enumerate(axnorm):
        for ii,mm in enumerate(mim):
            sim.set(norm, xyz[i])
            sim.set(" ".join([norm,mm]),xyzmm[i][ii])

    if material: sim.set("material", material)
    sim.set("override mesh order from material database",1); 
    sim.set("mesh order",zorder);
    

def quick_cylinder(sim, xyz,rx=1e-6,ry=1e-6,L=1e-6, xr=0,yr=0,zr=0, zorder=0, normal="z", material=None,name=None):
    sim.addcircle() #note that the circle is always instaned in the xy plane, normal to the z axis
    if name:
        sim.set("name",name)
    sim.set("first axis","x")
    sim.set("second axis","y")
    sim.set("third axis","z")
    if normal == "x":
        yr += 90
        
    if normal == "y":
        xr += 90
    sim.set("make ellipsoid",1)
    sim.set("radius",rx)
    sim.set("radius 2",ry)  
    sim.set("z span",L)
    sim.set("rotation 1",xr)
    sim.set("rotation 2",yr)
    sim.set("rotation 3",zr)     
    
    if material: sim.set("material", material)
    sim.set("override mesh order from material database",1); 
    sim.set("mesh order",zorder);
    
def L_primitive_old(sim,primitive="rect",xyz=None, x=0,y=0,z=0, Dx=1e-6, Dy=1e-6, Dz=1e-6, xmm = None, ymm=None,zmm = None,
                rx=1e-6, ry=1e-6, rz=1e-6, L=1e-6,xr=0,yr=0,zr=0,normal="z", material=None, zorder=0,bounds=None,name=None,group=None):
    if not xyz:
        xyz = (x,y,z)
    
    if primitive == "sphere":
        quick_sphere(sim,xyz,rx,ry,rz,zorder=zorder,material=material)
        xyz,minmax = radius_to_minmax(xyz, rx, ry, rz)
        adict = range_dict(xyz, *minmax)
        
    if primitive == "rect":
        if all(val == None for val in [xmm,ymm,zmm]):
            xyz,minmax = span_to_minmax(xyz, Dx,Dy,Dz)
        else:
            
            xyz,minmax = span_to_minmax(xyz, xmm,ymm,zmm)
        
        
        quick_cuboid(sim,xyz,*minmax)
        adict = range_dict(xyz, *minmax)
    if primitive == "cylinder":
        quick_cylinder(sim,xyz,rx=rx,ry=ry,L=L, xr=xr,yr=yr,zr=zr, zorder=zorder, normal=normal, material=material,name=name)
        adict = aabb_of_rotated_cylinder(center=(0,0,0), rx=rx, ry=ry, length=L, xr=xr, yr=yr, zr=zr)     
        
    if group:
        sim.addtogroup(group)
        
    if bounds:
        bounds.append(adict["rng"])
        
def L_nanowire(sim, NW, mat = None, seed_mat = None, zorder=0, axis_offset=(0,0,0),group=None):
    for sphere in NW.seed_list:
        sim.addsphere()
        
            
def L_roundedcube(sim, RC, material=None, zorder=0, axis_offset=(0,0,0), group=None,bounds=None):
    """
    Instantiate a RoundedCuboid in a Lumerical simulation using lumapi.
    
    sim: lumapi simulation object (e.g. FDTD(), varFDTD(), CHARGE(), etc.)
    RC: RoundedCuboid instance
    material: Lumerical material string
    zorder: drawing order
    axis_offset: (dx, dy, dz)
    group: name of group to insert objects into (string)
    """
    aox, aoy, aoz = axis_offset

    # -------------------------------------------------------------
    # CORE CUBES
    # -------------------------------------------------------------
    for cube in RC.c_cubes:
        cube_range = cube["range"]
        x = (cube_range["xmin"] + cube_range["xmax"]) / 2 + aox
        y = (cube_range["ymin"] + cube_range["ymax"]) / 2 + aoy
        z = (cube_range["zmin"] + cube_range["zmax"]) / 2 + aoz
        Dx = cube_range["xmax"] - cube_range["xmin"]
        Dy = cube_range["ymax"] - cube_range["ymin"]
        Dz = cube_range["zmax"] - cube_range["zmin"]

        L_primitive(sim,primitive="rect",xyz=(x,y,z), Dx=Dx, Dy=Dy, Dz=Dz,material=material,zorder=zorder,bounds=bounds,group=group)        

    # -------------------------------------------------------------
    # SPHERES
    # -------------------------------------------------------------
    for sphere in RC.r_spheres:
        loc = sphere["loc"]
        radius = sphere["radius"]

        x = loc["x"] + aox
        y = loc["y"] + aoy
        z = loc["z"] + aoz

        rx, ry, rz = radius["x"], radius["y"], radius["z"]
        L_primitive(sim,primitive="sphere",xyz=(x,y,z), rx=rx, ry=ry, rz=rz,material=material,zorder=zorder,bounds=bounds,group=group)      
        

    # -------------------------------------------------------------
    # CYLINDERS
    # -------------------------------------------------------------
    for cyl in RC.r_cylinders:
        loc = cyl["loc"]
        radius = cyl["radius"]
        axis_range = cyl["range"]
        norm = cyl["norm"]

        x = loc["x"] + aox
        y = loc["y"] + aoy
        z = loc["z"] + aoz

        rx = radius.get("x", 0)
        ry = radius.get("y", 0)
        rz = radius.get("z", 0)

        name = sim.addobject("cylinder")
        sim.setnamed(name, "x", x)
        sim.setnamed(name, "y", y)
        sim.setnamed(name, "z", z)

        sim.setnamed(name, "radius x", rx)
        sim.setnamed(name, "radius y", ry)
        sim.setnamed(name, "radius z", rz)

        sim.setnamed(name, "axis", norm)
        sim.setnamed(name, "z min", axis_range["zmin"])
        sim.setnamed(name, "z max", axis_range["zmax"])

        if material: sim.setnamed(name, "material", material)
        sim.setnamed(name, "z order", zorder)

        if group:
            sim.addtogroup(group)

def radius_to_minmax_oneax(loc, rad):
    """Returns [min,max] of a span, or [min,max] if span is already [min,max]"""
    if isinstance(rad, (list, tuple)):
        if len(rad) == 1:
            rad = rad[0]
        elif len(rad) == 2:
            return rad
    return(loc, [loc - rad, loc + rad])

def radius_to_minmax(xyz, rx,ry,rz):
    radii = [rx,ry,rz]
    minmax    = []
    for i,loc in enumerate(xyz):
        locnew,mm = radius_to_minmax_oneax(loc,radii[i])
        minmax.append(mm)
    return(xyz,minmax)


def span_to_minmax_oneax(loc, span):
    """Returns [min,max] of a span, or [min,max] if span is already [min,max]"""
    if isinstance(span, (list, tuple)):
        if len(span) == 1:
            span = span[0]
        elif len(span) == 2:
            locnew = np.mean(span)
            return(locnew,span)
    mm = [loc - span/2, loc + span/2]
    
    return(loc,mm)

def span_to_minmax(xyz,Dx,Dy,Dz):
    spans = [Dx,Dy,Dz]
    xyzn      = []
    minmax    = []
    for i,loc in enumerate(xyz):
        locnew,mm = span_to_minmax_oneax(loc,spans[i])
        xyzn.append(locnew)
        minmax.append(mm)
    
    return(xyzn,minmax)
    

def range_dict(xyz,xrange,yrange,zrange):
    adict = dict(loc={},rng = {"x":xrange,"y":yrange,"z":zrange}) 
    for i,dim in enumerate(["x","y","z"]):
        loc            = xyz[i]

        adict["loc"][dim], adict["rng"][dim] = span_to_minmax_oneax(loc, adict["rng"][dim]) 
    return(adict)


def determine_2D_3D_spans_normals(xyz,xrange,yrange,zrange,allow3D=True):
    adict = range_dict(xyz, xrange, yrange, zrange)
    
    # Determine which dimensions have non-zero span
    spans = {dim: adict["rng"][dim][1] - adict["rng"][dim][0] for dim in ["x","y","z"]}
    nonzero = [d for d in spans if spans[d] > 0]
    zero    = [d for d in spans if spans[d] == 0]

    # Determine 2D vs 3D
    if len(nonzero) not in [3,2]:
        raise ValueError(f"Invalid span combination: zero spans in {zero}. Provide either all 3 spans (3D) or exactly one zero span (2D).")
    if len(nonzero) == 3 and not allow3D:
        raise ValueError("Attempted to set 3 dimensions when only two are allowed. Please provide exactly one zero span (2D).")
    
    if len(nonzero) == 3:
        monitor_type = "3D"
        
    elif len(nonzero) == 2 and len(zero) == 1:
        monitor_type = "2D"
        for key,item in adict["rng"].items():
            if min(item) == max(item):
                normal = key
        monitor_type = "2D "+normal.upper()+"-normal"
    return(adict,monitor_type,[normal])
        
def L_addDFT(sim, name, group=None, xyz=(0,0,0), xrange=1, yrange=1, zrange=0,prop_dict={},bounds=None):
    """
    Add a 2D or 3D DFT monitor to the simulation.
    
    Two modes: span or range. You may provide either xrange=5 or xrange=[-2.5,2.5].
    
    Parameters:
        sim       : Lumerical simulation object
        name      : Name of the DFT monitor
        group     : Optional parent group (currently not used)
        xyz       : centre coordinates (tuple)
        xrange    : float or [min,max] for x
        yrange    : float or [min,max] for y
        zrange    : float or [min,max] for z
    """
    xyz = tuple(xyz)
    adict,monitor_type,skip = determine_2D_3D_spans_normals(xyz,xrange,yrange,zrange,allow3D=True)
    
    # Create the DFT monitor
    sim.adddftmonitor()
    sim.set("name",name)
    sim.addtogroup(group)
    sim.set("monitor type",monitor_type)
    
    for key,item in adict["loc"].items():
            sim.set(key,item)
            
    for key,item in adict["rng"].items():
        if key not in skip:
            sim.set(key+" min",item[0])
            sim.set(key+" max",item[1])
    select_and_set_props(sim, None, prop_dict)    
    if bounds:
        bounds.append(adict["rng"])
                
def L_addmovie(sim, name, group=None, xyz=(0,0,0), xrange=1, yrange=1, zrange=0,prop_dict={},bounds=None):
    """
    Add a 2D moviemonitor
    
    Two modes: span or range. You may provide either xrange=5 or xrange=[-2.5,2.5].
    Parameters:
        sim       : Lumerical simulation object
        name      : Name of  monitor
        group     : format: "model::group1::group2"
        xyz       : centre coordinates (tuple) - if using min/max, this can be anything
        xrange    : float or [min,max] for x
        yrange    : float or [min,max] for y
        zrange    : float or [min,max] for z
    """
    xyz = tuple(xyz)
    adict,monitor_type,skip = determine_2D_3D_spans_normals(xyz,xrange,yrange,zrange,allow3D=False)
    
    # Create the movie monitor
    sim.addmovie()
    sim.set("name",name)
    sim.addtogroup(group)
    
    sim.set("monitor type",monitor_type)
    
    for key,item in adict["loc"].items():
            sim.set(key,item)
            
    for key,item in adict["rng"].items():
        if key not in skip:
            sim.set(key+" min",item[0])
            sim.set(key+" max",item[1])
    select_and_set_props(sim, None, prop_dict)        
    if bounds:
        bounds.append(adict["rng"])
     

def L_addanalysis(sim, atype, name, group=None, xyz=(0,0,0), xrange=1, yrange=1, zrange=0,prop_dict={},bounds = None):
    """
    Add a 2D moviemonitor
    
    Two modes: span or range. You may provide either xrange=5 or xrange=[-2.5,2.5].
    Parameters:
        sim       : Lumerical simulation object
        name      : Name of  monitor
        group     : format: "model::group1::group2"
        xyz       : centre coordinates (tuple) - if using min/max, this can be anything
        xrange    : float or [min,max] for x
        yrange    : float or [min,max] for y
        zrange    : float or [min,max] for z
    """
    xyz = tuple(xyz)
    adict,monitor_type,skip = determine_2D_3D_spans_normals(xyz,xrange,yrange,zrange,allow3D=False)
    
    # Create the movie monitor
    sim.addmovie()
    sim.set("name",name)
    sim.addtogroup(group)
    
    sim.set("monitor type",monitor_type)
    
    for key,item in adict["loc"].items():
            sim.set(key,item)
            
    for key,item in adict["rng"].items():
        if key not in skip:
            sim.set(key+" min",item[0])
            sim.set(key+" max",item[1])
    
    select_and_set_props(sim, None, prop_dict)
    
    if bounds:
        bounds.append(adict["rng"])


def get_material_list(sim,substr=None,legacy=False):
    if not sim:
        legacy=True
        
    if legacy:
        Matlist = [ "Ag (Silver) - Johnson and Christy",
                    "Al (Aluminium) - CRC",
                    "Ag (Silver) - Palik (0-2um)",
                    "W (Tungsten) - Palik",
                    "InAs - Palik",
                    "TiO2 (Titanium Dioxide) - Sarkar",
                    "Cr (Chromium) - CRC",
                    "5CB - Li",
                    "Ge (Germanium) - Palik",
                    "Si3N4 (Silicon Nitride) - Luke",
                    "5PCH - Li",
                    "TiO2 (Titanium Dioxide) - Kischkat,"
                    "Ag (Silver) - CRC",
                    "PEC (Perfect Electrical Conductor),"
                    "Al2O3 - Palik",
                    "Al (Aluminium) - Palik",
                    "E44 - Li",
                    "W (Tungsten) - CRC",
                    "TiO2 (Titanium Dioxide) - Devore",
                    "Si3N4 (Silicon Nitride) - Phillip",
                    "Sn (Tin) - Palik",
                    "Cr (Chromium) - Palik",
                    "Fe (Iron) - Palik",
                    "MLC-6608 - Li",
                    "C (graphene) - Falkovsky (mid-IR)",
                    "Pt (Platinum) - Palik",
                    "Au (Gold) - CRC",
                    "Si (Silicon) - Palik",
                    "Fe (Iron) - CRC",
                    "SiO2 (Glass) - Palik",
                    "Si3N4 (Silicon Nitride) - Kischkat,"
                    "Pd (Palladium) - Palik",
                    "In (Indium) - Palik",
                    "Ni (Nickel) - CRC",
                    "MLC-9200-100 - Li",
                    "InP - Palik",
                    "MLC-9200-000 - Li",
                    "Ni (Nickel) - Palik",
                    "Cu (Copper) - CRC",
                    "Ge (Germanium) - CRC",
                    "etch",
                    "Au (Gold) - Palik",
                    "Ti (Titanium) - Palik",
                    "TiN - Palik",
                    "TiO2 (Titanium Dioxide) - Siefke",
                    "Ti (Titanium) - CRC",
                    "H2O (Water) - Palik",
                    "Au (Gold) - Johnson and Christy",
                    "GaAs - Palik",
                    "V (Vanadium ) - CRC",
                    "6241-000 - Li",
                    "Cu (Copper) - Palik",
                    "Ta (Tantalum) - CRC",
                    "E7 - Li",
                    "Rh (Rhodium) - Palik",
                    "Ag (Silver) - Palik (1-10um)",
                    "TL-216 - Li"]
    else:
        Matlist = sim.getmaterial().split("\n")
        
    if substr:
        # 1. If substr is a str → convert to list
        if isinstance(substr, str):
            substr_list = [substr]
        else:
            # 2. If it's already a list or tuple → use as-is
            substr_list = list(substr)

        # 3. Apply AND-filter:
        # Keep only materials that contain *all* substrings
        def match_all(mat):
            return all(s in mat for s in substr_list)

        Matlist = [mat for mat in Matlist if match_all(mat)]

    return(Matlist)

def loaddata(filepath):
    """
    Load a 3-column material file (wl, n, k).
    Accepts comma- or space-delimited formats automatically.
    Returns wl, n, k as numpy arrays.
    """

    # genfromtxt automatically detects commas or whitespace
    data = np.genfromtxt(
        filepath,
        comments="#",
        delimiter=None,      # auto-detect
        dtype=float,
        invalid_raise=False  # tolerate stray commas at end
    )

    # Remove empty rows that load as NaN
    data = data[~np.isnan(data).any(axis=1)]

    if data.shape[1] < 3:
        raise ValueError(f"{filepath} must contain at least 3 columns (wl, n, k).")
    cc = constants.c
    wl = np.array(data[:, 0])
    freq = 2 * np.pi * cc / (wl*1e-6)  # rad/s
    n  = np.array(data[:, 1])
    k  = np.array(data[:, 2])
    nik = n + 1j * k
    data_2col = np.column_stack((freq, nik))
    return wl, n, k,np.array(data_2col)

def import_materials_from_folder(sim,path,extlist=[".csv",".txt"],mat_type="Sampled 3D data",overwrite=True):
    filelist = [file for file in os.listdir(path) if any(ext in file for ext in extlist)]
    for file in filelist:
        fpath = os.path.join(path, file)
        material_name = file.split(".")[0]
        wl,n,k,data = loaddata(fpath)
        
        print(f"Importing material: {material_name}")

        # --- Create a new material in the simulation ---
        materiallist = sim.getmaterial().split("\n")
        if material_name in materiallist and overwrite==False:
            print(f"{material_name} already exists, skipping")
            return
        
        if material_name in materiallist and overwrite==True:
            sim.deletematerial(material_name)
        newmat = sim.addmaterial(mat_type)                   # creates a plain material
        sim.setmaterial(newmat,"name",material_name);
        sim.setmaterial(material_name, mat_type.lower(),data)
        
        print(f"Loaded {material_name} ({len(wl)} points)")

"""
HELPER SCRIPTS!!!
SCRIPTS BELOW ARE USED TO VERIFY FUNCTIONALITY OF THE CODE ABOVE, SUCH AS MATPLOTLIB PLOTTING OF CUBES, CYLINDER NANOWIRES, AND MORE!
"""
# ---------- CYLINDERS ----------
def check_ax(ax):
    try:
        bbox = ax.bbox
    except Exception as e:
        raise Exception("No axis provided") from e
        
def plot_cylinder(center, radius, height, ax=None, axis="z", resolution=20):
    # axis-aligned cylinder
    check_ax(ax)
        
    u = np.linspace(0, 2 * np.pi, resolution)
    h = np.linspace(-height/2, height/2, 2)
    U, H = np.meshgrid(u, h)
    if axis == "x":
        X = H + center["x"]
        Y = radius["y"] * np.cos(U) + center["y"]
        Z = radius["z"] * np.sin(U) + center["z"]
    elif axis == "y":
        X = radius["x"] * np.cos(U) + center["x"]
        Y = H + center["y"]
        Z = radius["z"] * np.sin(U) + center["z"]
    else:  # z-axis
        X = radius["x"] * np.cos(U) + center["x"]
        Y = radius["y"] * np.sin(U) + center["y"]
        Z = H + center["z"]
    ax.plot_surface(X, Y, Z, color="green", alpha=0.6)
    
def plot_cube(cube,ax=None):
    check_ax(ax)
    rng = cube["range"]
    x = [rng["xmin"], rng["xmax"]]
    y = [rng["ymin"], rng["ymax"]]
    z = [rng["zmin"], rng["zmax"]]

    # Create 2D surfaces for each face
    X, Y = np.meshgrid(x, y)
    ax.plot_surface(X, Y, z[0]*np.ones_like(X), color="cyan", alpha=0.3)  # bottom
    ax.plot_surface(X, Y, z[1]*np.ones_like(X), color="cyan", alpha=0.3)  # top

    Y, Z = np.meshgrid(y, z)
    ax.plot_surface(x[0]*np.ones_like(Y), Y, Z, color="cyan", alpha=0.3)  # left
    ax.plot_surface(x[1]*np.ones_like(Y), Y, Z, color="cyan", alpha=0.3)  # right

    X, Z = np.meshgrid(x, z)
    ax.plot_surface(X, y[0]*np.ones_like(X), Z, color="cyan", alpha=0.3)  # front
    ax.plot_surface(X, y[1]*np.ones_like(X), Z, color="cyan", alpha=0.3)  # back
    
def plot_sphere(sphere,resolution=20,ax=None):
    check_ax(ax)
    u = np.linspace(0, 2 * np.pi, resolution)
    v = np.linspace(0, np.pi, resolution)
    cx, cy, cz = sphere["loc"]["x"], sphere["loc"]["y"], sphere["loc"]["z"]
    rx, ry, rz = sphere["radius"]["x"], sphere["radius"]["y"], sphere["radius"]["z"]
    X = rx * np.outer(np.cos(u), np.sin(v)) + cx
    Y = ry * np.outer(np.sin(u), np.sin(v)) + cy
    Z = rz * np.outer(np.ones_like(u), np.cos(v)) + cz
    ax.plot_surface(X, Y, Z, color="red", alpha=0.6)
    
def plot_rounded_cuboid(CG,ax=None):
    check_ax(ax)
    for cube in CG.c_cubes:
        plot_cube(cube,ax=ax)

    # ---------- ROUNDING SPHERES ----------
    for sphere in CG.r_spheres:
        plot_sphere(sphere,resolution=20,ax=ax)
        
    for cyl in CG.r_cylinders:
        plot_cylinder(cyl["loc"], cyl["radius"], cyl["height"],ax=ax, axis=cyl["norm"])

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    ax.set_box_aspect([1,1,1])
    plt.show()

# --- Example usage ---
helper_functions = False
if helper_functions:
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    A = RoundedCuboid(rx=1, ry=1, rz=1, Dx=6, Dy=6, Dz=6, rx2=2, ry2=2, rz2=2)
    plot_rounded_cuboid(A,ax=ax)


def plot_Nanowire(NWG):       
    None
                                    



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 23 11:12:46 2020
@author: vidar
"""

import os
import glob,shutil
from setuptools import setup, find_packages
import ast

import distutils.cmd

try:
   from setupext_janitor import janitor
   CleanCommand = janitor.CleanCommand
except ImportError:
   CleanCommand = None

cmd_classes = {}
if CleanCommand is not None:
   cmd_classes['clean'] = CleanCommand
target_folder = "DMU" 
filelist = os.listdir(target_folder)
file_pathlist = [os.path.join(target_folder,file) for file in filelist  if ".py" in file]
TrgtScr = "DMU//utils.py"


class CleanCommand(distutils.cmd.Command):
    """
    Our custom command to clean out junk files.
    """
    description = "Cleans out junk files we don't want in the repo"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cmd_list = dict(
            DS_Store="find . -name .DS_Store -print0 | xargs -0 rm -f;",
            pyc="find . -name '*.pyc' -exec rm -rf {} \;",
            empty_dirs="find ./pages/ -type d -empty -delete;"
        )
        for key, cmd in cmd_list.items():
            os.system(cmd)



"""
Run commands to find modules from the given script file
"""
modules = set()
def visit_Import(node):
    for name in node.names:
        modules.add(name.name.split(".")[0])

def visit_ImportFrom(node):
    # if node.module is missing it's a "from . import ..." statement
    # if level > 0 it's a "from .submodule import ..." statement
    if node.module is not None and node.level == 0:
        modules.add(node.module.split(".")[0])
    
def module_filter(modules,exceptions):
    mod2 = list(modules)
    for module in mod2:
         if module in []:
             modules.remove(module)

    return(list(modules))
        

module_exceptions = [file.split(".py")[0].replace(os.path.join(target_folder,""),"") for file in filelist] + ["mpl_toolkits","json","time","sys","os","tkinter","collections","csv","pickle","torch", "kornia", "torchvision"]       


node_iter = ast.NodeVisitor()
node_iter.visit_Import = visit_Import
node_iter.visit_ImportFrom = visit_ImportFrom


with open(TrgtScr) as f:
    node_iter.visit(ast.parse(f.read()))


with open("README.md", "r") as fh:
    long_description = fh.read()


installREQ = [module_filter(modules,module_exceptions)+["docutils>=0.3"]]
installREQ = [[item for item in installREQ[0] if item not in ["logging"]]]
extras_require = { "alldeps": [
                    "torch",
                    "torchvision",
                    "kornia",
                ],
                }
setup(
    name="DMU",
    version="0.4.1",
    packages=find_packages(),
    scripts=[],

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires =installREQ,
    extras_require   = extras_require,
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.txt", "*.json"],
        # And include any *.msg files found in the "hello" package, too:
        "hello": ["*.msg"],
    }, 
    # metadata to display on PyPI
    author="Atli Vidar Már FLodgren",
    author_email="vidar@flodgren.com",
    description="This package is used to store commonly used functions on pip for the sake of easy pulling for other code. Features utilities for IV characterisation, lumerical data analysis and geometry instanciation, SEM stitching and other tools.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="hello world example examples",
    url="https://github.com/DeltaMod/DMU",   # project home page, if any
    #project_urls={
    #    "Bug Tracker": "https://bugs.example.com/HelloWorld/",
    #    "Documentation": "https://docs.example.com/HelloWorld/",
    #    "Source Code": "https://code.example.com/HelloWorld/",
    #},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha"]

    # could also include long_description, download_url, etc.
)



import os
#%% Importing and executing logging
import logging


try:
    from . custom_logger import get_custom_logger
    logger = get_custom_logger("DMU_SEMUTILS")
    # Importing plot tools 
    from . utils_utils import *
    from . plot_utils import *
    from . utils import kwarg_aliasing 
    
except:
    from custom_logger import get_custom_logger
    logger = get_custom_logger("DMU_SEMUTILS")
    # Importing plot tools 
    from utils_utils import *
    from plot_utils import *
    from utils import kwarg_aliasing 
    print("Loading plot_utils packages locally, since root folder is the package folder")

    
import tifffile
import svgwrite
from PIL import Image,ImageEnhance,ImageOps, ImageFilter
import base64
from io import BytesIO
import numpy as np 
import os 
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import subprocess
import tempfile
import torch
import glob
import urlopen
import cv2 as cv
import sys 
import PIL
try:
    import kornia as K
    import kornia.feature as KF
except ImportError:
    raise ImportError(
       "Torch and Kornia support requires `pip install DMU[alldeps]`")
    
import ssl
import certifi
from matplotlib.patches import Rectangle

def svg_to_pil(svgdrawing, inkscape_path, mode="L"):
    """
    Convert an svgwrite.Drawing object to a PIL image using Inkscape CLI.
    Works reliably on Windows by using temporary files.
    """
    # Convert SVG to bytes
    svg_bytes = svgdrawing.tostring().encode("utf-8")

    # Create a temporary SVG file
    fd_svg, tmp_svg_name = tempfile.mkstemp(suffix=".svg")
    os.close(fd_svg)  # Close immediately so Inkscape can write
    with open(tmp_svg_name, "wb") as f:
        f.write(svg_bytes)

    # Create a temporary PNG filename
    fd_png, tmp_png_name = tempfile.mkstemp(suffix=".png")
    os.close(fd_png)

    # Run Inkscape to export PNG
    cmd = [
        inkscape_path,
        tmp_svg_name,
        "--export-type=png",
        f"--export-filename={tmp_png_name}"
    ]
    subprocess.run(cmd, check=True)

    # Read PNG into PIL
    pil_img = Image.open(tmp_png_name).convert(mode)

    # Clean up temporary files
    os.remove(tmp_svg_name)
    os.remove(tmp_png_name)
    return(pil_img)

def ANY_Image_Enhance(PIL_im,brightness=None,contrast=None,sharpness=None,expand_range=True):
    
    try:
        PIL_im = PIL_im.convert("L")
    except:
        if PIL_im.dtype == np.float32 or PIL_im.dtype == np.float64:
            img_array_uint8 = (PIL_im * 255).astype(np.uint8)
        else:
            img_array_uint8 = PIL_im
        # Convert to PIL Image
        PIL_im = Image.fromarray(img_array_uint8)
        
    if expand_range:
        pixvals = np.array(PIL_im)

        
        pixvals = ((pixvals - pixvals.min()) / (pixvals.max()-pixvals.min())) * 255
        PIL_im = Image.fromarray(pixvals.astype(np.uint8))
    
    if brightness:
        enhancer = ImageEnhance.Brightness(PIL_im)
        PIL_im = enhancer.enhance(brightness)
        
    if contrast:
        enhancer = ImageEnhance.Contrast(PIL_im)
        PIL_im = enhancer.enhance(contrast)
        
    if sharpness:
        enhancer = ImageEnhance.Sharpness(PIL_im)
        PIL_im = enhancer.enhance(sharpness)
    
    
    return(PIL_im)

def SEM_Annotation_Finder(image_path):
    """
    Provide image path, and you shall be given the crop matrix used in IMAGE.crop([left,upper,right,lower]). 
    ----------
    image_path : r'path\to\file.tif'
        

    crop_matrix = [left,upper,right,lower]
    
    If no annotation is found, then this crop matrix 

    """
    # Load image and convert to grayscale
    im = Image.open(image_path).convert('L')  # Convert to grayscale
 
    # Convert image to a numpy array
    im_array = np.array(im)
 
    # Apply a binary threshold (0 = black, 255 = white)
    binary_image = np.where(im_array == 0, 0, 255).astype(np.uint8)
 
    for row_idx, row in enumerate(binary_image):
        black_pixel_count = np.sum(row == 0)
        if black_pixel_count > 0.9 * len(row):  # 90% of the row is black
            l_crop = im.height - row_idx
            if l_crop%2!=0:
                l_crop+=1
            imcrop = [0,0,0,im.height - row_idx]    
            return(imcrop)
    
    imcrop = [0,0,0,0]
    return(imcrop)

          
def split_crop_bounds_evenly(length,newlength,offset=0):
    crop = length-newlength
    if crop%2==0:
        cropl=cropr=crop/2
    else:
        cropl=(crop-1)/2
        cropr = crop-cropl
    return(cropl+offset,length-cropr+offset)

def SEM_Strip_Banner_And_Enhance(image_path,filterdict={}):
    """
    filterdict_raw = dict(brightness=1,
                          contrast=1,
                          sharpness=1,
                          expand_range=True,
                          resampling="bicubic",
                          blur=dict(type=None,radius=0),
                          strip_banner=True,
                          tweak_aspect=[1,1],
                          force_aspect=False,
                          imcrop=[0,0,0,0],
                          crop_rescale=[0,0]
                          rotation=0,
                          delta_offset=[0,0])

    """
    filterdict_raw = dict(brightness=1,
                          contrast=1,
                          sharpness=1,
                          expand_range=True,
                          resampling="bicubic",
                          blur=dict(type=None,radius=0),
                          strip_banner=True,
                          tweak_aspect=[1,1],
                          force_aspect=False,
                          imcrop=[0,0,0,0],
                          crop_rescale=[0,0],
                          rotation=0,
                          delta_offset=[0,0],
                          resize=1)
    
    filterdict_raw.update({k: v for k, v in filterdict.items() if k in filterdict_raw})
    FD = filterdict_raw
    sem_metadata = SEM_get_metadata(image_path)
    
    FD["resampling"] = {"nearest":Image.Resampling.NEAREST,"bicubic":Image.Resampling.BICUBIC,"bilinear":Image.Resampling.BILINEAR,"lancoz":Image.Resampling.LANCZOS,"box":Image.Resampling.BOX}[FD["resampling"].lower()]
        
    with Image.open(image_path) as im:
        buffer = BytesIO()
        imformat = im.format            
        
        if FD["strip_banner"] == True:
            annocrop= SEM_Annotation_Finder(image_path)        
            im = im.crop((annocrop[0],annocrop[1],im.width-annocrop[2],im.height-annocrop[3]))
            
        
        og_w,og_h = (im.width, im.height)
        
        if FD["tweak_aspect"] != [1,1]:
            im = im.resize((int(im.width*FD["tweak_aspect"][0]),int(im.height*FD["tweak_aspect"][1])),resample=FD["resampling"])    
            og_w,og_h = (im.width, im.height)
        
        if FD["force_aspect"]!=False:
            og_w,og_h = find_nearest_aspect_dim(im.width,im.height,FD["force_aspect"])
        
        if FD["imcrop"] != [0,0,0,0]:
            im = im.crop((FD["imcrop"][0],FD["imcrop"][1],im.width- FD["imcrop"][2],im.height-FD["imcrop"][3]))
           
            
        if FD["rotation"] != 0:
            
            im = im.rotate(FD["rotation"], expand=True)
            
            
        if FD["force_aspect"] != False:
            xd,yd = find_nearest_aspect_dim(im.width,im.height,FD["force_aspect"])
            xcrop = split_crop_bounds_evenly(im.width, xd,offset=FD["delta_offset"][0])
            ycrop = split_crop_bounds_evenly(im.height, yd,offset=FD["delta_offset"][1])
            im = im.crop((xcrop[0],ycrop[0],xcrop[1],ycrop[1]))
        
        
            
        if FD["crop_rescale"] == True:
            pix_rescale = 1/(og_w/im.width)
            im = im.resize((og_w,og_h),resample=FD["resampling"])
        else:
            pix_rescale = 1
        

            
        if FD["resize"] != None:
            rszm = FD["resize"]
            im = im.resize((int(im.width*rszm),int(im.height*rszm)),resample=FD["resampling"])
            pix_rescale *=1/(rszm)
        else:
            rszm = 1
            
        if len(FD.keys())!=0:
            im = ANY_Image_Enhance(im,**filterdict)
        
        # --- NEW: Apply blur ---
        blur_info = FD.get("blur", {})
        if blur_info.get("type") is not None:
            blur_type = blur_info["type"].lower()
            radius = blur_info.get("radius")
            if blur_type == "gaussian":
                im = im.filter(ImageFilter.GaussianBlur(radius=radius))
            elif blur_type == "box":
                im = im.filter(ImageFilter.BoxBlur(radius))
            elif blur_type == "average":
                # BoxBlur with radius approximates averaging
                im = im.filter(ImageFilter.BoxBlur(radius))
                
        im.format = imformat

        
    return(im.convert("L"),sem_metadata)
    
    
def SEM_get_metadata(image_path):

    with tifffile.TiffFile(image_path) as tif:
        og_w,og_h = (np.shape(tif.asarray())[1], np.shape(tif.asarray())[0])
        sem_metadata = tif.sem_metadata
        if sem_metadata == None:
            
            try:
                if tif.fei_metadata["System"]["SystemType"] == 'Nova NanoLab':                
                    sem_metadata = tif.fei_metadata["EScan"]
                    sem_metadata["sv_serial_number"] = ["Serial Code",tif.fei_metadata["System"]["SystemType"]]
                    sem_metadata["ap_image_pixel_size"] = ["ap_image_pixel_size",sem_metadata["PixelWidth"],"m"]
                    pix_size_string = "ap_image_pixel_size"
                    sem_metadata["ap_stage_at_t"] = ["rotation",tif.fei_metadata["Stage"]["SpecTilt"]]
                
            except:
                print("SEM MODEL NOT IMPLEMENTED!!! FIX IMPORTER")
        elif "SUPRA 35-29-41" in sem_metadata['sv_serial_number'][1]:
            pix_size_string = "ap_image_pixel_size"
        elif "Gemini" in sem_metadata['sv_serial_number'][1]:
            pix_size_string = "ap_image_pixel_size"
        elif "1560-95-96" in sem_metadata["sv_serial_number"][1]:
            pix_size_string = "ap_width"
        else:
            pix_size_string = "ap_pixel_size"
    sem_metadata["pix_size_string"] = pix_size_string    
    
    ##While we're at it, we also calculate the pixel size in absolute units. We do this so we can reference it later in scaling topics. pix rescale does not need to exist here either.
    OOM_raw = {"pm":1e-12,"nm":1e-9,"um":1e-6,'µm':1e-6,"mm":1e-3,"m":1e+0}
    OOM = {normalize_unit_key(k): v for k,v in OOM_raw.items()}
    #Find the scale parameters from the image in question: 
    
    if pix_size_string != "ap_width":
        pix_size = sem_metadata[pix_size_string][1] * OOM[sem_metadata[pix_size_string][2]]
        
    elif pix_size_string == "ap_width":
        pix_size = sem_metadata[pix_size_string][1] * OOM[sem_metadata[pix_size_string][2]]/og_w
      
    sem_metadata["pix_size"] = pix_size      
    return(sem_metadata)


def normalize_unit_key(s):
    #This function fixes issues with how mu is displayed in certain contexts
    if not isinstance(s, str):
        return s
    try:
        # Decode from bytes in case of mojibake
        s = s.encode('latin1').decode('utf-8')
    except (UnicodeEncodeError, UnicodeDecodeError):
        pass
    # Standardize common ASCII alternatives
    s = s.replace('u', 'µ') if s.lower().startswith('um') else s
    return s
            
def SEM_Scalebar_Generator(image_path, svg_output, scalebar_style = {},txt_style={}, imcrop=[0,0,0,0], savefile=True, resize=None, 
                           remove_annotation=True, resampling="bicubic",force_aspect=False,delta_offset=[0,0],rotation=0,crop_rescale=True,tweak_aspect=[1,1],filterdict={},recalculate_stroke_width=True,
                           sem_metadata=None,imformat_override="tiff"):
    
    
    """
    Example image_path = 'DFR1-HE_BR204.tif' (or any literal string address)
    you can also give it an image, and sem_metadata if you want to avoid image_path stuff. 
    Example svg_output = 'output.svg'
    If svg_output is set to "Auto" then the name image_path+anno.svg is used
    
    scalebar_style and txt_style refer to the possible parameters we pass in a dict. Below are two examples that contain something in EVERY field possible currently.
    
    scalebar_style = {"frame":True,
                      "framepad":[30,2],
                      "stroke_width":4, %this should be calculated from a percentage of the image width
                      "stroke_style":"line",
                      "bar_color":"white",
                      "frame_color":"black",
                      "frame_opacity":0.6,
                      "location":"lower right",
                      "location_padding":[0.03,0.05],
                      "bar_ratio":[1/6,1/40]}
    
    txt_style={"font":"Arial",
               "fontsize":"Auto",
               "font_weight":"normal",
               "font_style":"normal",
               "text_decoration":"none",
               "color":"white"}
    
    Example use 1:
    dwg = SEM_Scalebar_Generator(image_path, svg_output, scalebar_style=scalebar_style,txt_style=txt_style, imcrop=[0,0,0,0])
    
    Example use 2:
        for file in [f for f in os.listdir() if f.endswith(".tif")]:
        dwg = SEM_Scalebar_Generator(image_path, "Auto", scalebar_style=scalebar_style,txt_style=txt_style, imcrop=[0,0,0,0])
    """
    
    """
    Defining the default scalebar and text settings
    """
    if sem_metadata == None:
        sem_metadata = SEM_get_metadata(image_path)
    pix_size_string = sem_metadata["pix_size_string"]
    
    
    def scalebar_style_dictgen(frame=None,framepad=[2,2],stroke_width=4,stroke_style="line",
                                 bar_color="white",frame_color="black",frame_opacity=1.0,
                                 location="lower left",location_padding=[0.03,0.05],
                                 bar_ratio=[1/6,1/40]):
            
        
        return({"frame":frame,"framepad":framepad,"stroke_width":stroke_width,"stroke_style":stroke_style,"bar_color":bar_color,
                "frame_color":frame_color,"location":location,"bar_ratio":bar_ratio,"location_padding":location_padding,"frame_opacity":frame_opacity})
    
    def text_style_dictgen(im,font_family="Arial",fontsize="Auto",font_fraction=1/20,font_weight="normal",font_style="normal",text_decoration="none",color="black"):
        if fontsize == "Auto":
            fontsize = int(im.height/20)
        if fontsize == "fraction":
            fontsize= int(im.height * font_fraction)
        
        return({"font_family":font_family,"fontsize":fontsize,"font_weight":font_weight,"font_style":font_style,"text_decoration":text_decoration,"color":color})
    
    
    resampling = {"nearest":Image.Resampling.NEAREST,"bicubic":Image.Resampling.BICUBIC,"bilinear":Image.Resampling.BILINEAR,"lancoz":Image.Resampling.LANCZOS,"box":Image.Resampling.BOX}[resampling.lower()]
    pix_rescale = 1
    
    if type(image_path) == str:

        if svg_output == "Auto":
            svg_output = image_path.split(".")[0]+"_anno.svg"
            
        with Image.open(image_path) as im:
            buffer = BytesIO()
            imformat = im.format            
            
            if remove_annotation == True:
                annocrop= SEM_Annotation_Finder(image_path)        
                im = im.crop((annocrop[0],annocrop[1],im.width-annocrop[2],im.height-annocrop[3]))
                
            
            og_w,og_h = (im.width, im.height)
            
            if tweak_aspect != [1,1]:
                im = im.resize((int(im.width*tweak_aspect[0]),int(im.height*tweak_aspect[1])),resample=resampling)    
                og_w,og_h = (im.width, im.height)
            
            if force_aspect!=False:
                og_w,og_h = find_nearest_aspect_dim(im.width,im.height,force_aspect)
            
            if imcrop != [0,0,0,0]:
                im = im.crop((imcrop[0],imcrop[1],im.width-imcrop[2],im.height-imcrop[3]))
               
                
            if rotation != 0:
                
                im = im.rotate(rotation, expand=True)
                
                
            if force_aspect != False:
                xd,yd = find_nearest_aspect_dim(im.width,im.height,force_aspect)
                xcrop = split_crop_bounds_evenly(im.width, xd,offset=delta_offset[0])
                ycrop = split_crop_bounds_evenly(im.height, yd,offset=delta_offset[1])
                im = im.crop((xcrop[0],ycrop[0],xcrop[1],ycrop[1]))
            
            
                
            if crop_rescale == True:
                pix_rescale = 1/(og_w/im.width)
                im = im.resize((og_w,og_h),resample=resampling)
            else:
                pix_rescale = 1
            
    
                
            if resize != None:
                rszm = resize
                im = im.resize((int(im.width*rszm),int(im.height*rszm)),resample=resampling)
                pix_rescale *=1/(rszm)
            else:
                rszm = 1
                
            if len(filterdict.keys())!=0:
                im = ANY_Image_Enhance(im,**filterdict)
            
            im.format = imformat
    
            im.save(buffer, format=im.format)
            #im.save(buffer, format=im.format)  # Save the image in its original format to a buffer
            img_str = base64.b64encode(buffer.getvalue()).decode('utf-8')  # Encode image as base64
            
    else:
        im = image_path
        buffer = BytesIO()
        
        if type(im) == Image.Image:
            if im.format == None:
                im.format = imformat_override
            im.format.replace(".","")
        if im.format.lower() in ["tif"]:
            im.format = "tiff"
        im.save(buffer, format=im.format)
        img_str = base64.b64encode(buffer.getvalue()).decode('utf-8')
        
    txt  = text_style_dictgen(im,**txt_style)    

    
    
    sbar = scalebar_style_dictgen(**scalebar_style)
    if recalculate_stroke_width:
        sbar["stroke_width"] *=im.width/1000 
    
    txt_orig = txt.copy()
    sbar_orig = sbar.copy()
    
    OOM_raw = {"pm":1e-12,"nm":1e-9,"um":1e-6,'µm':1e-6,"mm":1e-3,"m":1e+0}
    OOM = {normalize_unit_key(k): v for k,v in OOM_raw.items()}
    #Find the scale parameters from the image in question: 
    
    if pix_size_string != "ap_width":
        pix_size = sem_metadata[pix_size_string][1] * OOM[sem_metadata[pix_size_string][2]]*pix_rescale
        
    elif pix_size_string == "ap_width":
        pix_size = sem_metadata[pix_size_string][1] * OOM[sem_metadata[pix_size_string][2]]*pix_rescale/og_w
      
    rtilt = np.radians(sem_metadata['ap_stage_at_t'][1])
    rrot = np.radians(rotation)
    
    if abs(sem_metadata['ap_stage_at_t'][1]) >= 5 and abs(rotation) > 10:
            L = pix_size/np.sin(rtilt)
            pix_size = L*np.cos(np.arcsin(np.cos(rrot)*np.sin(rtilt))) 
            #pix_size = np.sqrt(L**2*(np.cos(rrot)**2*np.cos(rtilt)**2 + np.sin(rrot)**2)) #L**2 - L**2*np.cos(rtilt)**2*np.sin(rrot)**2
    
    #Define bar length and height. Note that setting the height to zero removes it

    sbar["bar_length_target"] = int(im.width*sbar["bar_ratio"][0]) * pix_size 
    sbar["bar_height"] = int(im.height*sbar["bar_ratio"][1]) 
    
    def find_nearest_scale_bar(target_length):
        """
        We list all multipliers, powers of 10 we will consider, and calculate allowed values.
        """
        # List of allowed multipliers
        multipliers = np.array([1, 2, 5, 10, 20, 50, 100, 200, 500])
        scale_dict_raw = {"1e+09":"Gm","1e+06":"Mm","1e+03":"km","1e+00":"m","1e-03":"mm","1e-06":"µm","1e-09":"nm","1e-12":"pm"}
        OOMS = np.array([1e+9, 1e+6, 1e+3, 1e+0, 1e-3, 1e-6, 1e-9])    
        scale_dict = {normalize_unit_key(k): v for k,v in scale_dict_raw.items()}
        allowed_values = (multipliers[:, None] * OOMS).flatten()
        powers_array = np.tile(OOMS, len(multipliers))
        
        if target_length == 0:
            return 0, 1  # Return 0 for scale bar and 1 as power of 10 for simplicity
        
        # Find the index of the nearest value in allowed_values
        nearest_index = np.abs(allowed_values - target_length).argmin()
        
        # Get the nearest scale bar and its corresponding power of ten
        nearest_scale_bar = allowed_values[nearest_index]
        
        nearest_power_of_ten = powers_array[nearest_index]
        
        len_string = str(np.round(nearest_scale_bar/nearest_power_of_ten,1))
        if len_string.endswith(".0"):
            len_string = len_string.replace(".0","")
          
        len_string +=scale_dict["{:.0e}".format(nearest_power_of_ten)]
          
        return(nearest_scale_bar, nearest_power_of_ten, len_string)
    
    
    
    def scalebar_generation(svg_output,im,scalebar_style,txt_style,pix_size):
        #Draw direction and location determination
        sbar = scalebar_style; txt = txt_style

        best_scale, best_oom, len_string = find_nearest_scale_bar(sbar["bar_length_target"])
        
        bar_length = best_scale/pix_size
        
        
        if "lower" in sbar["location"]:
            y_coord = int(im.height - im.height*sbar["location_padding"][1])
            draw_dir_y = -1
        elif "upper" in sbar["location"]:
            y_coord = int(im.height*sbar["location_padding"][1])
            draw_dir_y = 1
        else:
            y_coord = int(im.height/2)
        if "left" in sbar["location"]:
            draw_dir_x = 1
            x_coord = int(im.width*sbar["location_padding"][0])
        elif "right" in sbar["location"]: 
            x_coord = int(im.width-im.width*sbar["location_padding"][0])
            draw_dir_x = -1
        else:
            x_coord = int(im.width/2)
            draw_dir_x = 0
        draw_dir = [draw_dir_x,draw_dir_y]
        location = [x_coord,y_coord]
        bar_middle_x = draw_dir[0] * bar_length/2 + location[0]
        bar_start = tuple(location)
        bar_end   = tuple([draw_dir[0] * bar_length + location[0],location[1]] )
        
        textloc = tuple([bar_middle_x,bar_start[1] + int(sbar["bar_height"]*1.2) * draw_dir[1]])
        
        # Create an SVG drawing with svgwrite
        dwg = svgwrite.Drawing(svg_output, profile='full', size=(im.width, im.height))
        
        # Add the raster image as base64 inside the SVG
        dwg.add(dwg.image(href=f"data:image/{im.format.lower()};base64,{img_str}",
                          insert=(0, 0), size=(im.width, im.height)))
        
        if sbar["frame"] != None:
            y_min = int(textloc[1] - txt["fontsize"]/1.5   - sbar["framepad"][1] - sbar["stroke_width"]*2)
            y_max = int(location[1] + sbar["bar_height"] + sbar["framepad"][1]+sbar["stroke_width"]*2)
            box_dxy = (np.abs(bar_start[0] - bar_end[0])+ sbar["framepad"][0] + sbar["stroke_width"]*2,y_max-y_min + sbar["framepad"][1])
            box_xy = (bar_middle_x - box_dxy[0]/2,np.mean([y_max,y_min])-box_dxy[1]/2) 
            
            dwg.add(dwg.rect(
                            insert=box_xy,  # Bottom-left corner of the rectangle
                            size=box_dxy,  # Width and height of the rectangle
                            fill=sbar["frame_color"],  # Fill color of the background (you can choose any color)
                            stroke="none",  # Optional stroke for the rectangle
                            stroke_width=0,
                            fill_opacity=sbar["frame_opacity"]
                        ))
        
        # Draw a vector line on top of the image
        dwg.add(dwg.line(start=bar_start, end=bar_end, 
                         stroke=sbar["bar_color"], stroke_width=sbar["stroke_width"]))
        
        if sbar["bar_height"] != 0:
            dwg.add(dwg.line(start=(bar_start[0],bar_start[1] + sbar["bar_height"]), end=(bar_start[0],bar_start[1] - sbar["bar_height"]), 
                             stroke=sbar["bar_color"], stroke_width=sbar["stroke_width"]))
            dwg.add(dwg.line(start=(bar_end[0],bar_end[1] + sbar["bar_height"]), end=(bar_end[0],bar_end[1] - sbar["bar_height"]), 
                             stroke=sbar["bar_color"], stroke_width=sbar["stroke_width"]))
        
        dwg.add(dwg.text(len_string,insert=textloc,text_anchor="middle", font_size=txt["fontsize"], font_family=txt["font_family"], fill=txt["color"]))

        # Save the SVG
    
        return(dwg)
    dwg = scalebar_generation(svg_output,im,sbar,txt,pix_size)
    if savefile == True:
        dwg.save()
    return({"svg":dwg,"im":im,"sbar":sbar_orig,"txt":txt_orig})
    
#%%
def find_nearest_aspect_dim(width, height, ratio):
    if width > height*ratio:
        width = int(height*ratio + 0.5)
    else:
        height = int(width/ratio + 0.5)
    return (width, height)



def SEM_Create_Insert(image_overview, inserts, filename="Auto",path="", scalebar_style = {},txt_style={},force_aspect=4/3,filterdict={},imcrop=[0,0,0,0],rotation=0):
    """
    This function wraps the SEM_Scalebar_Generator and creates inserts based on the provided information in the dicts. 
    The list of dicts must contain {"path":path\to\image,"size":1/N,"framecolor":CMAPformat, "framing":[x,y,w,h],"location":[1,1],"loc_type":"grid"}
    If you provide a size 1/N image, then the image_overview is scaled by Nx, and the number of choices you get in the grid mode becomes (N-1)
    In short, you get a 3x3 grid with a 1/2 size image, and a 7x7 grid from a 1/4 size, and a 15x15 from a 1/8 size and the coordinates for each becomes exactly as you expect.
    If you choose "loc_type":"pixels" then you simply provide the centre pixel that you want
    If you choose "figure_fraction" then you use fractions. These methods do not account for overlap.
    
    Parameters
    ----------
    image_overview : TYPE
        DESCRIPTION.
    inserts_dict : TYPE
        DESCRIPTION.
    scalebar_style : TYPE, optional
        DESCRIPTION. The default is {}.
    txt_style : TYPE, optional
        DESCRIPTION. The default is {}.

    Returns
    -------
    None.

    """
    def inserts_dict_generator(path=None,size=1/2,frame_color=plt.get_cmap("tab20c")(5), stroke_width=4, framing=[0,0,100,100],location=[1,1],loc_type="grid",delta_offset=[0,0],force_aspect=4/3,imcrop=[0,0,0,0],filterdict={},rotation=0):
        return({"path":path,"size":size,"frame_color":frame_color, "stroke_width":stroke_width, "framing":framing,"location":location,"loc_type":loc_type,"delta_offset":delta_offset,"imcrop":imcrop, "force_aspect":force_aspect,"filterdict":filterdict,"rotation":rotation})
    
    inserts = [inserts_dict_generator(**insert) for insert in inserts] #refactor inserts to adhere to formatting. Can't really be automatic, but will help I think
    
    with Image.open(image_overview) as im:
        width = im.width
        smallwidth = im.width*np.min([insert["size"] for insert in inserts])
        
    insert_widths  = []
    scalefactors   = []
    for insert in inserts:
        with Image.open(insert["path"]) as im:
            insert_widths.append(im.width)
            insert["base_scale_factor"] = smallwidth/im.width
            scalefactors.append(smallwidth/im.width)
    
    if np.min(scalefactors) < 1:
        #Scale up overview, don't scale anything that has same scalefactor. Scale up things with greater scale factors
        overview_scale = 1/np.min(scalefactors)
        for insert in inserts:
            insert["scale_factor"] = np.min(scalefactors) / insert["base_scale_factor"]  
        
    
    
    if np.min(scalefactors) >= 1:
        #Scale up everything else, don't scale anything that has same scalefactor. Scale up things with greater scale factors
        overview_scale = 1
        for insert in inserts:
            insert["scale_factor"] = insert["base_scale_factor"]
            
    if filename == "Auto":
        image_overview_name = image_overview.split("\\")[-1].split(".")[0]+"_combined_inserts.svg"
        image_overview_path = os.path.join(path,image_overview_name)
    
    else:
        image_overview_name = filename.split(".")[0] + ".svg"
        image_overview_path = os.path.join(path,image_overview_name)
    
    
    dwg = SEM_Scalebar_Generator(image_overview, image_overview_path, scalebar_style=scalebar_style,txt_style=txt_style, imcrop=imcrop,force_aspect=force_aspect,delta_offset=[0,0],resize=overview_scale,filterdict=filterdict,rotation=rotation)
    
    
    scalebar_style = dwg["sbar"]
    txt_style= dwg["txt"]
    insert_svglist = []
    for insert in inserts:
        s_factor     = insert["scale_factor"] 
        #location grid for the insert's sizing 
        points = np.linspace(0, 1, int((1/insert["size"]*2-1)**2))
        # Create the grid using np.meshgrid
        x_grid, y_grid = np.meshgrid(points*dwg["im"].width, points*dwg["im"].height)
        # Stack the x_grid and y_grid to create an array of tuples (x, y)
        cgrid = np.dstack((x_grid, y_grid)) # this is accessed as "row,col" 
        
        isbs = scalebar_style
     
        isbs["framepad"]         = [scalebar_style["framepad"][0],scalebar_style["framepad"][1]] 
        isbs["stroke_width"]     = scalebar_style["stroke_width"]
        isbs["bar_ratio"]        = [scalebar_style["bar_ratio"][0]/insert["size"]*0.8,scalebar_style["bar_ratio"][1]/insert["size"]]
        isbs["location_padding"] = [scalebar_style["location_padding"][0]/insert["size"],scalebar_style["location_padding"][1]/insert["size"]] 
        
        itxt = txt_style 
        if itxt["fontsize"] == "Auto":
            itxt["fontsize"] = int(dwg["im"].height/10)
        insert_path = insert["path"]
       
        insert_name = filename.split(".")[0]+"_"+insert_path.split("\\")[-1].split(".")[0]+"_insert.svg"
        insert_newpath = os.path.join(path,insert_name)
        insert_svg  = SEM_Scalebar_Generator(insert_path, insert_newpath, scalebar_style=isbs,txt_style=itxt, imcrop=insert["imcrop"],
                                                           force_aspect=insert["force_aspect"],delta_offset=insert["delta_offset"],resize=insert["scale_factor"],filterdict=insert["filterdict"],rotation=insert["rotation"],recalculate_stroke_width=False)
        
        insert_svglist.append(insert_svg)
        img_loc = cgrid[insert["location"][0],insert["location"][1]]
        img_loc = (img_loc[0],img_loc[1])
        if img_loc[0] >= dwg["im"].width/2:
            ilx=-isbs["stroke_width"]
        elif img_loc[0] < dwg["im"].width/2:
            ilx=isbs["stroke_width"]
        else:
            ilx = 0
            
        if img_loc[1] >= dwg["im"].height/2:
            ily=-isbs["stroke_width"]
        elif img_loc[1] < dwg["im"].height/2:
            ily=isbs["stroke_width"]
        else:
            ily = 0
            
        img_loc = (img_loc[0] + ilx, img_loc[1]+ily)
        
        href = 'file:///' + insert_newpath.replace('\\', '/')
        image = dwg["svg"].image(href=href, insert=img_loc, size=(insert_svg["im"].width, insert_svg["im"].height))
        dwg["svg"].add(image)
        
        if insert["framing"] is not None:
            # Calculate rx and ry as 5% of the width of the rectangle
            rx = ry = 0*0.05 * insert_svg["im"].width
        
            # Create the first rectangle with rounded corners
            rect1 = dwg["svg"].rect(
                insert=(img_loc[0], img_loc[1]),  # Bottom-left corner of the rectangle
                size=(insert_svg["im"].width, insert_svg["im"].height),  # Width and height of the rectangle
                fill="none",  # Fill color of the background (you can choose any color)
                stroke=mcolors.to_hex(insert["frame_color"]),  # Optional stroke for the rectangle
                stroke_width=insert["stroke_width"] * dwg["im"].width / 1000,
                rx=rx,
                ry=ry
            )
            dwg["svg"].add(rect1)
        
            # Create the second rectangle with rounded corners
            rect2 = dwg["svg"].rect(
                insert=(insert["framing"][0] - insert["framing"][2] / 2, insert["framing"][1] + insert["framing"][3] / 2),  # Bottom-left corner of the rectangle
                size=(insert["framing"][2], insert["framing"][3]),  # Width and height of the rectangle
                fill="none",  # Fill color of the background (you can choose any color)
                stroke=mcolors.to_hex(insert["frame_color"]),  # Optional stroke for the rectangle
                stroke_width=insert["stroke_width"] * dwg["im"].width / 1000,
                rx=rx,
                ry=ry
            )
            dwg["svg"].add(rect2)
        
            # Create a clipping path using the first rectangle
            clip_path = dwg["svg"].defs.add(dwg["svg"].clipPath(id="clip"))
            clip_path.add(dwg["svg"].rect(
                insert=(img_loc[0], img_loc[1]),
                size=(insert_svg["im"].width, insert_svg["im"].height),
                rx=rx,
                ry=ry
            ))
        
            # Apply the clipping path to the image
            image['clip-path'] = f'url(#{clip_path.get_id()})'
    # Save the SVG file
    dwg["svg"].save()
    
    
    return(dwg)
    
"""    
scalebar_style = {"frame":True,
                  "framepad":[30,2],
                  "stroke_width":4,
                  "stroke_style":"line",
                  "bar_color":"white",
                  "frame_color":"black",
                  "frame_opacity":0.6,
                  "location":"lower right",
                  "location_padding":[0.03,0.05],
                  "bar_ratio":[1/6,1/40]}

txt_style={"font":"Arial",
           "fontsize":34,
           "font_weight":"normal",
           "font_style":"normal",
           "text_decoration":"none",
           "color":"white"}
"""

def export_svg_to_png(inkscape_path, svg_file, output_png, width=None, height=None, dpi=None):
    """
    Export SVG to PNG using Inkscape command-line interface (CLI).
    
    Parameters:
    - inkscape_path: Path to the Inkscape executable
    - svg_file: Path to the input SVG file
    - output_png: Path to the output PNG file
    - width: Optional, desired width of the PNG
    - height: Optional, desired height of the PNG
    - dpi: Optional, desired DPI of the PNG
    """
    cmd = [inkscape_path, svg_file, '--export-type=png', '--export-filename=' + output_png]

    if width:
        cmd += ['--export-width=' + str(width)]
    if height:
        cmd += ['--export-height=' + str(height)]
    if dpi:
        cmd += ['--export-dpi=' + str(dpi)]

    subprocess.run(cmd)


####
"""
STITCHING FUNCTIONS AND CODE FOR IMAGE MERGING
"""
def stitcher_add_rectangles(ax, centers, widths, heights, colours):
    """
    Overlay outline rectangles on an axis.
    
    Parameters:
        ax : matplotlib.axes.Axes
            Axis to draw on
        centers : list of (x, y)
            Centres of rectangles
        widths : list or float
            Width(s) of rectangles
        heights : list or float
            Height(s) of rectangles
        colours : list
            List of colours to cycle through
    """
    for i, (cx, cy) in enumerate(centers):
        w = widths[i] if isinstance(widths, (list, np.ndarray)) else widths
        h = heights[i] if isinstance(heights, (list, np.ndarray)) else heights
        color = colours[i % len(colours)]
        rect = Rectangle(
            (cx - w/2, cy - h/2),
            w, h,
            edgecolor=color,
            facecolor='none',
            linewidth=2
        )
        ax.add_patch(rect)
        
def stitcher_enhance_for_matching(img):
    img_uint8 = (img*255).astype(np.uint8)
    clahe = cv.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))
    img_eq = clahe.apply(img_uint8)
    return img_eq.astype(np.float32)/255.0


def stitcher_enforce_similarity(H_affine):
    """
    Given a 2x3 affine matrix, returns the nearest similarity transform:
    translation + rotation + uniform scale, no shear.
    """

    A = H_affine[:, :2]     # 2×2 linear part
    t = H_affine[:, 2:]     # 2×1 translation

    # SVD decomposition A = U Σ V^T
    U, S, Vt = np.linalg.svd(A)

    # Uniform scale = average of the two singular values
    scale = S.mean()

    # Pure rotation matrix
    R = U @ Vt

    # Fix reflection if determinant is -1
    if np.linalg.det(R) < 0:
        U[:,-1] *= -1
        R = U @ Vt

    # Reconstruct similarity transform
    A_sim = scale * R
    H_sim = np.hstack([A_sim, t])

    return H_sim.astype(np.float32)

#Currenty unused, unfortunately
def stitcher_refine_similarity_confident(mkpts0, mkpts1, H_init, min_inliers=10, max_error=3.0):
    """
    Refine an initial 2x3 transform H_init into a similarity transform.
    Returns H_refined or None if refinement fails/confidence is too low.
    """
    # Estimate affine2D for similarity
    H_affine, mask = cv.estimateAffine2D(mkpts1, mkpts0, method=cv.RANSAC, ransacReprojThreshold=max_error)

    if H_affine is None:
        return None

    # Convert to similarity
    H_sim = stitcher_enforce_similarity(H_affine)

    # Compute inliers / reprojection error
    mkpts1_proj = cv.transform(mkpts1.reshape(-1,1,2), H_sim)[:,0,:]
    errors = np.linalg.norm(mkpts1_proj - mkpts0, axis=1)
    inliers = np.sum(errors < max_error)

    if inliers < min_inliers:
        return None

    return H_sim

def sem_stitcher(farpics,nearpics,filename="Auto",filesave = True, seam_dict=None, matching_model=None, scalebar_style=None,txt_style=None, inkscape_path=r"C:\Program Files\Inkscape\bin\inkscape.exe"):
    """
    farpic/nearpic can either be a filelist, or a glob string (no * needed). Choose what fits best
    """
    ### Initialise Default values for seam handling, and scalebar styles.
    scalebar_style_defaults = {"frame":True,
                               "framepad":[30,2],
                               "stroke_width":6,
                               "stroke_style":"line",
                               "bar_color":"white",
                               "frame_color":"black",
                               "frame_opacity":0.6,
                               "location":"lower right",
                               "location_padding":[0.03,0.05],
                               "bar_ratio":[1/6*1.05,1/40]}

    txt_style_defaults={"font_family":"Arial",
                        "fontsize":"Auto",
                        "font_weight":"normal",
                        "font_style":"normal",
                        "text_decoration":"none",
                        "color":"white"}
    
    seam_dict_defaults = dict(transition=400, gamma=3.0)
    seam_aliases = dict(transition=["transition"], gamma=["gamma"])
    if not seam_dict:
        seam_dict = {}
    if not txt_style:
        txt_style = {}
    if not scalebar_style:
        scalebar_style= {}
        
    seam_dict      = kwarg_aliasing(seam_dict,seam_dict_defaults,aliases=seam_aliases)
    scalebar_style = kwarg_aliasing(scalebar_style,seam_dict_defaults,aliases=None)
    txt_style      = kwarg_aliasing(txt_style,seam_dict_defaults,aliases=None)
        
    #Determining if glob or filelist has been used.
    if type(farpics) == str:
        farpics = sorted(glob.glob(farpics+"*"))
     
    if type(nearpics) == str:
        nearpics = sorted(glob.glob(nearpics+"*"))
   
    farpics_firstname  = farpics[0]
    if nearpics:
        nearpics_firstname = nearpics[0]
    else:
        nearpics_firstname = ""
        
    IMGD     = dict(farpic = [],nearpic=[],filenames=[],sem_metadata_f=[],sem_metadata_c=[])
    
    
    # ---- MODEL CONFIG ---- #
    DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    print("Using device:", DEVICE)
    
    if matching_model is None:
        matcher = KF.LoFTR(pretrained='outdoor').to(DEVICE)
    else:
        # Load a torch hub model by name
        # example: matching_model = 'resnet18'
        model_name = matching_model.lower()
        if model_name == "resnet18":
            ssl._create_default_https_context = lambda: ssl.create_default_context(cafile=certifi.where())
            matcher = torch.hub.load('pytorch/vision:v0.15.2', 'resnet18', pretrained=True).to(DEVICE)
        else:
            raise ValueError(f"Unknown matching_model '{matching_model}'")
        
    plt.close("all")
    plt.ion()
    
    
    # ---- LOAD AND PREPROCESS IMAGES ---- #
    for i,f in enumerate(farpics):
        farpic  = farpics[i]
        
        if "tif" in farpic:
            img_proc, sem_metadata = SEM_Strip_Banner_And_Enhance(farpic, filterdict=dict(expand_range=False))
            img_proc = np.asarray(img_proc)
            
            if img_proc is None:
                continue
        
        try:    
            nearpic = nearpics[i]
            if "tif" in nearpics[i]:
                imgn_proc, sem_metadata2 = SEM_Strip_Banner_And_Enhance(nearpic, filterdict=dict(expand_range=False))
                sbimg =  SEM_Scalebar_Generator(imgn_proc, "temp.svg", scalebar_style=scalebar_style,txt_style=txt_style, remove_annotation=False, sem_metadata=sem_metadata2)
                sbimg = svg_to_pil(sbimg["svg"], inkscape_path) 
        except:
            imgn_proc = None; sem_metadata2 = None; sbimg = None; 
            
        IMGD["farpic"].append(img_proc)
        IMGD["nearpic"].append(sbimg)
        IMGD["sem_metadata_f"].append(sem_metadata)
        IMGD["sem_metadata_c"].append(sem_metadata2)
        IMGD["filenames"].append(farpic)
        
          
    # ---- ACCUMULATE AFFINE TRANSFORMS ---- #
    H_list = [np.eye(2,3, dtype=np.float32)]  # store 2x3 matrices for warpAffine
  
    for i in range(len(IMGD["farpic"])-1):
        img1 = stitcher_enhance_for_matching(IMGD["farpic"][i])
        img2 = stitcher_enhance_for_matching(IMGD["farpic"][i+1])
        
        t1 = torch.from_numpy(img1).unsqueeze(0).unsqueeze(0).float().to(DEVICE)
        t2 = torch.from_numpy(img2).unsqueeze(0).unsqueeze(0).float().to(DEVICE)
        
        with torch.no_grad():
            out = matcher({"image0": t1, "image1": t2})
            mkpts0 = out["keypoints0"].cpu().numpy()
            mkpts1 = out["keypoints1"].cpu().numpy()
        
        print(f"Pair {i}-{i+1}: matches = {len(mkpts0)}")
        
        # Skip pairs with too few matches
        if len(mkpts0) < 4:
            print(f"Skipping pair {i}-{i+1}: insufficient matches")
            H_list.append(H_list[-1].copy())
            continue
        
        # Step 1: translation-only
        H_translation, _ = cv.estimateAffinePartial2D(mkpts1, mkpts0, method=cv.RANSAC)
        
        if H_translation is None:
            H_list.append(H_list[-1].copy())
            continue
        MIN_MATCHES_FOR_REFINEMENT = 10
    
        if len(mkpts0) < MIN_MATCHES_FOR_REFINEMENT:
            H_final = H_translation  # safe fallback
        else:
            # try affine refinement
            H_affine, mask = cv.estimateAffine2D(mkpts1, mkpts0, method=cv.RANSAC)
            if H_affine is not None:
                H_refined = stitcher_enforce_similarity(H_affine)
                # accept only if it improves mean reprojection error
                mkpts1_refined = cv.transform(mkpts1.reshape(-1,1,2), H_refined)[:,0,:]
                errors_refined = np.linalg.norm(mkpts1_refined - mkpts0, axis=1)
                mean_error_refined = errors_refined.mean()
        
                mkpts1_trans = cv.transform(mkpts1.reshape(-1,1,2), H_translation)[:,0,:]
                errors_translation = np.linalg.norm(mkpts1_trans - mkpts0, axis=1)
                mean_error_translation = errors_translation.mean()
        
                if mean_error_refined < mean_error_translation:
                    # also check rotation magnitude
                    angle = np.arctan2(H_refined[1,0], H_refined[0,0]) * 180/np.pi
                    if abs(angle) < 5:  # arbitrary max rotation in degrees
                        H_final = H_refined
                    else:
                        H_final = H_translation
                else:
                    H_final = H_translation
            else:
                H_final = H_translation
        
        # Accumulate in homogeneous coordinates
        H_last_h = np.vstack([H_list[-1], [0,0,1]])
        H_final_h = np.vstack([H_final, [0,0,1]])
        H_accum_h = H_last_h @ H_final_h
        H_list.append(H_accum_h[:2])
    
    # ---- BUILD PANORAMA CANVAS ---- #
    all_corners = []
    
    for img, H in zip(IMGD["farpic"], H_list):
        h, w = img.shape
        corners = np.array([[0,0],[w,0],[w,h],[0,h]], dtype=np.float32)
        warped = cv.transform(corners.reshape(-1,1,2), H)
        all_corners.append(warped.reshape(-1,2))
            
    all_pts = np.vstack(all_corners)
    x_min, y_min = np.floor(all_pts.min(axis=0)).astype(int)
    x_max, y_max = np.ceil(all_pts.max(axis=0)).astype(int)
    
    W = x_max - x_min
    Hh = y_max - y_min
    print("Panorama size:", W, "x", Hh)
    # ---- DETERMINE IMAGE CENTRES and square sizes ---- #
    centers = []
    for img, H in zip(IMGD["farpic"], H_list):
        h_img, w_img = img.shape
    
        H_h = np.vstack([H, [0, 0, 1]])
    
        shift_h = np.eye(3, dtype=np.float32)
        shift_h[0, 2] = -x_min
        shift_h[1, 2] = -y_min
    
        H_shift_h = shift_h @ H_h
        center_img = np.array([[w_img / 2, h_img / 2]],dtype=np.float32).reshape(-1, 1, 2)
    
        center_panorama = cv.perspectiveTransform(center_img, H_shift_h)
    
        cx, cy = center_panorama[0, 0]
        centers.append((cx, cy))
        
    scale_difference = [] 
    for i,img in enumerate(IMGD["sem_metadata_f"]):
        try:
            scale_difference.append(IMGD["sem_metadata_c"][i]["pix_size"]/IMGD["sem_metadata_f"][i]["pix_size"])
        except:
            scale_difference.append(None)
        
            
    h_img, w_img = IMGD["farpic"][-1].shape
    square_size = int(min(h_img, w_img) * (scale2 / scale1))
    
    panorama_sum = np.zeros((Hh, W), dtype=np.float32)
    panimg = PIL.Image.fromarray(panorama_sum)
    weight_mask = np.zeros((Hh, W), dtype=np.float32)
    # Compute square size from last image
    
    #We now take the IMGD["farpic"], and send them through the image enhancer with a filterdict.
    filterdict = dict(brightness=1,contrast=1.1,sharpness=1.1,expand_range=True)
    IMGD["farpic"] = [ANY_Image_Enhance(im,**filterdict) for im in IMGD["farpic"]]
    sbimg =  SEM_Scalebar_Generator(IMGD["farpic"][-1], "temp.svg", scalebar_style=scalebar_style,txt_style=txt_style, remove_annotation=False, sem_metadata=IMGD["sem_metadata_f"][-1])
    
    IMGD["farpic"][-1] = svg_to_pil(sbimg["svg"], inkscape_path) 
    
    #%%
    #Now we import and convert the matching farpic
    
    #%%
    tbc = dmp.get_tab20bc(grouping="pairs",output="list")[0::2]
    warped_imgs = []
    weights = []

    TRANSITION = seam_dict["transition"]
    gamma = seam_dict["gamma"]

    for img, H in zip(IMGD["farpic"], H_list):
        if isinstance(img, PIL.Image.Image):
            img = np.array(img)

        H_h = np.vstack([H, [0,0,1]])
        shift_h = np.eye(3, dtype=np.float32)
        shift_h[0,2] = -x_min
        shift_h[1,2] = -y_min
        H_shift = (shift_h @ H_h)[:2]

        warped = cv.warpAffine(img, H_shift, (W, Hh))
        mask = (warped > 0).astype(np.uint8)

        dist = cv.distanceTransform(mask, cv.DIST_L2, 5)
        dist = np.minimum(dist, TRANSITION) / TRANSITION
        dist = dist ** gamma

        warped_imgs.append(warped.astype(np.float32))
        weights.append(dist.astype(np.float32))

    weight_sum = np.sum(weights, axis=0) + 1e-8
    weights = [w / weight_sum for w in weights]

    panorama_avg = np.zeros_like(warped_imgs[0], dtype=np.float32)
    for warped, w in zip(warped_imgs, weights):
        panorama_avg += warped * w

    
    
    if panorama_avg.shape[0] > panorama_avg.shape[1]:
        #This means we have a vertical image, and therefore the panorama should cover 2 rows
        rcstart = [0,1]
        nrows = 2
        ncols = int(np.ceil(len(IMGD["farpic"])/nrows))+1
        ifrac = 2.5
        Gspec = plt.GridSpec(nrows, ncols,width_ratios=[1/ifrac]+[(1-1/ifrac)/(ncols-1) for val in range(ncols-1)])  
        PanGS = Gspec[:,0]
    
    
    else:
        #This means we have a horzontal image, and as such the panorama should cover 2 cols
        rcstart = [1,0] 
        ncols = 2
        nrows = int(np.ceil(len(IMGD["farpic"])/ncols))+1
        Gspec = plt.GridSpec(nrows, ncols,width_ratios=[0.3 for val in range(ncols)])  
        PanGS = Gspec[0,:]
    gslist = []
    for n in range(rcstart[0],nrows):
        for m in range(rcstart[1],ncols):
            gslist.append(Gspec[n,m])    
    
    # Create figure
    fig = plt.figure(figsize=(4 * ncols*1.5, 4 * nrows))
    
    # Create the panorama axis
    ax_pan = fig.add_subplot(PanGS)
    
    # Create the image axes using your gslist
    ax_ins = [fig.add_subplot(gs) for gs in gslist]
    
    fig.subplots_adjust(wspace=0, hspace=0)
    #dwg = SEM_Scalebar_Generator(item["sourcepath"], svgpath, scalebar_style=scalebar_style,txt_style=txt_style, imcrop=item["imcrop"],resize=2,delta_offset=item["delta_offset"], crop_rescale=True,force_aspect=4/3,tweak_aspect=tweak_aspect,rotation=item["rotation"],filterdict=item["filterdict"],savefile=False)
    
    
    figsolo, ax = plt.subplots(figsize=(12,10))
    ax.imshow(panorama_avg, cmap='gray')
    ax.set_axis_off()
    ax.set_title("SEM Panorama (Weighted Average)")
    
    # Define rectangle sizes
    w_rect = w_img * (scale1 / scale2)
    h_rect = h_img * (scale1 / scale2)
    
    # Add rectangles
    stitcher_add_rectangles(ax, centers, w_rect, h_rect, tbc)
    
    #Put Together the composite
    ax_pan.imshow(panorama_avg,cmap="gray")
    ax_pan.set_axis_off()
    for i,img in enumerate(IMGD["nearpic"]):
        ax_ins[i].imshow(img, cmap='gray')
        ax_ins[i].set_axis_off()
        
        # Add a colored rectangle around the image
        rect = Rectangle(
            (0, 0), img.width, img.height,               # x, y, width, height
            linewidth=6,                # border thickness
            edgecolor=tbc[i],         # border color from your list
            facecolor='none'            # transparent fill
        )
        ax_ins[i].add_patch(rect)
    
    stitcher_add_rectangles(ax_pan, centers, w_rect, h_rect, tbc)
    if filename == "Auto":
        filename = "stitch_"+farpics_firstname+nearpics_firstname
    if savefig == True:
        fig.savefig(filename)
    
    plt.show()





